# ../virtuals/__init__.py

"""Retrieves and stores virtual offsets from Linux binaries."""

# =============================================================================
# >> IMPORTS
# =============================================================================
# Python Imports
#   MMap
from mmap import ACCESS_READ
from mmap import mmap
#   OS
from os import O_RDWR
from os import open as _open
#   RegularExpression
from re import match
#   Struct
from struct import calcsize
from struct import unpack
#   Warnings
from warnings import warn

# Site-Package Imports
#   Configobj
from configobj import ConfigObj
#   Path
from path import Path
#   Virtuals
from virtuals.options import virtual_option_parser


# =============================================================================
# >> GLOBAL VARIABLES
# =============================================================================
# Get the current file
_file = Path(__file__)

# Get the configuration
config = ConfigObj(_file.parent.joinpath(_file.parent.namebase + '.ini'))

# Get the start paths from the configuration
BINARY_DIR = Path(config['BINARY_DIR'])
SAVE_DIR = Path(config['SAVE_DIR'])


# =============================================================================
# >> FUNCTIONS
# =============================================================================
def main():
    """Find virtual offsets for the provided command-line options."""
    # Get the options provided
    options, args = virtual_option_parser.parse_args()

    # Were no games provided?
    if options.games is None:

        # Use all games
        games = {game.namebase for game in BINARY_DIR.dirs()}

    # Otherwise, use the games provided
    else:
        games = set(options.games.split(','))

    # Get the binaries provided
    binaries = options.binaries

    # Loop through all games
    for game in games:

        # Get the game's binary path
        game_path = BINARY_DIR.joinpath(game)

        # If the path is not found, raise a warning
        if not game_path.isdir():
            warn('Path not found: {0}'.format(game_path))
            continue

        # Were binaries provided?
        if binaries is not None:

            # Loop through the given binaries
            for binary in binaries:

                # Get the binary's path
                binary_path = game_path.joinpath(binary + '.so')

                # If the file is not found, attempt to check for _srv filename
                if not binary_path.isfile():
                    binary_path = game_path.joinpath(binary + '_srv.so')

                    # If neither filename was found, raise a warning
                    if not binary_path.isfile():
                        warn('Binary "{0}" not found for game "{1}"'.format(
                            binary, game))
                        continue

                # Find all virtuals for the current binary
                find_virtuals(game, binary_path)

        # If not binaries were provided, find
        #   virtuals for all binaries for the game
        else:
            for file in game_path.files('*.so'):
                find_virtuals(game, file)


def find_virtuals(game, file):
    """Find the virtual offsets for the given file and save them."""
    # Get the path to the save directory
    game_save_path = SAVE_DIR.joinpath(game, file.namebase.replace('_srv', ''))

    # Remove all current files in the save directory to refresh
    if game_save_path.isdir():
        for filepath in game_save_path.walkfiles():
            filepath.remove()
        game_save_path.removedirs()

    # Create the save directory
    game_save_path.makedirs()

    # Thanks to L'In20Cible for this section.
    # I really don't know fully how this part works, so it is uncommented.
    mm = mmap(_open(file, O_RDWR), 0, access=ACCESS_READ)
    hdr = unpack('16s2H5I6H', mm[:calcsize('16s2H5I6H')])
    if hdr[0][:4] != b'\x7fELF':
        mm.close()
        return
    sections = list()
    for index in range(hdr[12]):
        offset = hdr[6] + (hdr[11] * index)
        sections.append(unpack('10I', mm[offset:offset + hdr[11]]))
    offset = sections[hdr[13]][4]
    names = mm[offset:offset + sections[hdr[13]][5]]
    for section in sections:
        name = names[section[0]:names.find(b'\x00', section[0])]
        if b'dyn' not in name:
            if section[1] == 2:
                data = section
            elif section[1] == 3:
                strings = section
    strings = mm[strings[4]:strings[4] + strings[5]]
    symbols = dict()
    for index in range(int(data[5] / data[9])):
        offset = calcsize('3L2BH') * index
        hdr = unpack('3L2BH', mm[data[4]:data[4] + data[5]][
            offset:offset + calcsize('3L2BH')])
        symbols[hdr[1]] = (
            strings[hdr[0]:strings.find(b'\x00', hdr[0])],) + hdr
    for data in symbols.values():
        matches = match(b'_ZTV([a-zA-Z0-9_]+)', data[0])
        if not matches:
            continue
        matches = match(b'^([0-9]+)([a-zA-Z0-9_]+)', matches.group(1))
        if not matches:
            continue
        name = matches.group(2)[:int(matches.group(1))]
        with game_save_path.joinpath(
                name.decode() + '.txt').open('w') as open_file:
            for index in range(1, int(data[3] / 4)):
                addr = unpack('L', mm[data[2] + 4 + (
                    index * 4):data[2] + 4 + (index * 4) + 4])[0]
                if not addr:
                    name = '<pure virtual>'
                else:
                    name = symbols.get(
                        addr, (b'<unresolved>', ))[0].decode()
                open_file.write('{0}\t{1}\n'.format(index - 1, name))
    mm.close()
